// charge le greffon "Yeti" (pour la transformée en ondelette "à trous") :
require, "yeti.i";
wavelet = symlink_to_name("yeti_wavelet");  // on fait un alias "dynamique"

// initialise les options graphiques par défaut :
pldefault, style="boxed.gs", marks=0, width=1, palette="stern.gp",
  maxcolors=240, legends=0, dpi=100;

// Pré-charge des définitions de fonction très utiles pour ce TP
TPDIR = setup_package();
require, TPDIR + "img.i";        // manipulation d'image (tableau 2D en Yorick)
require, TPDIR + "fft_utils.i";  // fonctions utilitaires pour la FFT
require, TPDIR + "utils.i";      // fonctions utilitaires générales
require, TPDIR + "plot.i";       // fonctions de tracé
require, TPDIR + "histo.i";      // fonctions utilitaires pour histogrammes


// fonction d'affichage d'image simplifiée
func show(img, win, cmin=, cmax=)
/* DOCUMENT show, img;
         or show, img, win;

      Simplified version to display an image.  Take care of resetting the
      graphics window.
     
   SEE ALSO: pli, limits, logxy, fma, window.
 */
{
  // sélectionne la fenêtre
  if (! is_void(win)) window, win;

  // efface les graphismes courants et réinitialise la fenêtre
  fma;
  logxy, 0, 0;
  limits, square=1;
  //limits;
  dims = dimsof(img);
  pli, img, 0.5, 0.5, dims(2) + 0.5, dims(3) + 0.5, cmin=cmin, cmax=cmax;
}

func cg(a, b, xinit=, gtol=, maxiter=, extra=, verb=)
/* DOCUMENT cg(a, b)

     Conjugate gradient routine. A is the funtion to call to compute
     the product of the "LHS matrix" of the problem with a vector
     given in argument.  B is the "RHS vector".

     Keyword XINIT can be used to specify an initial solution to start
     with. Be default, the initial solution is zero.

     Keyword GTOL (1E-5 by default) is the relative gradient tolerance for
     convergence.  The conjugate gradient iterations are performed until the
     squared Euclidean norm of the residual is less or equal GTOL times the
     squared Euclidean norm of the initial residuals.

     Keyword MAXITER can be used to set the maximum number of iterations which
     is twice the number of unknown by default.

     Keyword EXTRA can be used to pass additional parameters to the function
     A.  If EXTRA is non-nil, then the function is called as A(V, EXTRA)
     whenever the product of the "LHS matrix" with the vector V needs to be
     computed.


   SEE ALSO:
 */
{
  x = (is_void(xinit) ? array(double, dimsof(b)) : xinit);
  if (is_void(gtol)) gtol = 1e-5;
  if (is_void(maxiter)) maxiter = 2*numberof(x);
  no_extra = is_void(extra);
  r = b - (no_extra ? a(x) : a(x, extra));
  k = 0;
  while (k < maxiter) {
    rho = sum(r*r);
    if (k == 0) {
      tol = gtol*rho;
      p = r;
    } else if (rho <= tol) {
      break;
    } else {
      p = r + (rho/rho0)*p;
    }
    if (verb) {
      /*
       * r = b - A.x ==> A.x = b - r
       * f(x) = 1/2 x'.A.x - x'.b + c
       *      = 1/2 x'.(b - r) - x'.b + c
       *      = c - 1/2 x'.(b + r)
       */
      penalty = -0.5*sum(x*(b + r));
      write, format="%4d  %22.15E\n", k, penalty;
    }
    q = (no_extra ? a(p) : a(p, extra));
    alpha = rho/sum(p*q);
    x = x + alpha*p;
    r = r - alpha*q;
    rho0 = rho;
    ++k;
  }
  return x;
}

func ifft(z)
/* DOCUMENT ifft(z)
     Compute the inverse FFT or Z, assuming that Z is Hermitian (i.e. is
     the FFT of a real array.
   SEE ALSO: fft.
 */
{
  return (1.0/numberof(z))*double(fft(z, -1));
}

func wavelet_filter(w, t)
/* DOCUMENT wavelet_filter(w, t)
     This function reconstructs a filtered version of a signal by hard
     thresholding of the wavelet decomposition of the signal.  W is the
     wavelet decomposition of the signal and T is an array of thresholds, one
     for each wavelet plane.

     Example:
       wavelet_filter(wavelet(img, 3), [5.0, 11.0, 12.5]);
     
   SEE ALSO: wavelet.
 */
{
  dims = dimsof(w);
  nw = dims(0) - 1; // number of wavelet planes
  nt = numberof(t); // number of thresholds
  out = w(.., 0);
  for (k = 1; k <= nw; ++k) {
    tmp = w(.., k);
    if (k <= nt) {
      out += (abs(tmp) > t(k))*tmp;
    } else {
      out += tmp;
    }
  }
  return out;
}

func DtD(x)
/* DOCUMENT DtD(x)
     returns the result of D'.D.x where D is a (multi-dimensional)
     finite difference operator and D' is its transpose.
     
   SEE ALSO:
 */
{
  if (identof(x) > Y_DOUBLE) {
    error, "bad data type";
  }
  dims = dimsof(x);
  r = array(double, dims); // to store the result
  ndims = dims(1); // number of dimensions (up to 10 in Yorick)
  p = 1:-1;
  q = 2:0;
  if (ndims >= 1) {
    if (dims(2) >= 2) {
      dx = x(dif,..);
      r(q,..) += dx;
      r(p,..) -= dx;
    }
    if (ndims >= 2) {
      if (dims(3) >= 2) {
        dx = x(,dif,..);
        r(,q,..) += dx;
        r(,p,..) -= dx;
      }
      if (ndims >= 3) {
        if (dims(4) >= 2) {
          dx = x(,,dif,..);
          r(,,q,..) += dx;
          r(,,p,..) -= dx;
        }
        if (ndims >= 4) {
          if (dims(5) >= 2) {
            dx = x(,,,dif,..);
            r(,,,q,..) += dx;
            r(,,,p,..) -= dx;
          }
          if (ndims >= 5) {
            if (dims(6) >= 2) {
              dx = x(,,,,dif,..);
              r(,,,,q,..) += dx;
              r(,,,,p,..) -= dx;
            }
            if (ndims >= 6) {
              if (dims(7) >= 2) {
                dx = x(,,,,,dif,..);
                r(,,,,,q,..) += dx;
                r(,,,,,p,..) -= dx;
              }
              if (ndims >= 7) {
                if (dims(8) >= 2) {
                  dx = x(,,,,,,dif,..);
                  r(,,,,,,q,..) += dx;
                  r(,,,,,,p,..) -= dx;
                }
                if (ndims >= 8) {
                  if (dims(9) >= 2) {
                    dx = x(,,,,,,,dif,..);
                    r(,,,,,,,q,..) += dx;
                    r(,,,,,,,p,..) -= dx;
                  }
                  if (ndims >= 9) {
                    if (dims(10) >= 2) {
                      dx = x(,,,,,,,,dif,..);
                      r(,,,,,,,,q,..) += dx;
                      r(,,,,,,,,p,..) -= dx;
                    }
                    if (ndims >= 10) {
                      if (dims(11) >= 2) {
                        dx = x(,,,,,,,,,dif,..);
                        r(,,,,,,,,,q,..) += dx;
                        r(,,,,,,,,,p,..) -= dx;
                      }
                      if (ndims > 10) {
                        error, "too many dimensions";
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return r;
}

/*---------------------------------------------------------------------------*/
